import { Schema } from "../schema/classes/Schema";
import Api, { RQLParams, ResultSet, CurrentUser } from "./Api";
import {
  NativeTransactionBackend,
  Transaction,
  TransactionBackend,
  TransactionResult,
} from "./Transaction";

/**
 * Class used to communicate with a CubicWeb instance's API.
 */
export default class Client {
  private api: Api;
  private transactionBackend: TransactionBackend;

  /**
   * Class used to communicate with a CubicWeb instance's API.
   *
   * @param apiUrl The API url. It should not include the API version as it will be automatically added.
   * Example: if you can retrieve the schema with https://example.com/api/v1/schema,
   * use https://example.com/api as `apiUrl`.
   */
  constructor(
    apiUrl: string,
    {
      transactionBackend: transactionBackend,
    }: { transactionBackend?: TransactionBackend }
  ) {
    this.api = new Api(apiUrl);
    if (transactionBackend === undefined) {
      this.transactionBackend = new NativeTransactionBackend();
    } else {
      this.transactionBackend = transactionBackend;
    }
  }

  get apiUrl() {
    return this.api.apiUrl;
  }

  /**
   * Executes the given query and waits for the response.
   *
   * @param query The RQL query to send
   * @param params The additional parameters for the request
   * @returns A promise with the request's result set
   */
  async execute(query: string, params: RQLParams = {}): Promise<ResultSet> {
    return await this.api.execute(query, params);
  }

  /**
   * Retrieves the CubicWeb instance's JSON schema.
   *
   * @returns The CubicWeb instance's JSON schema.
   */
  async schema() {
    const schema = await this.api.schema();
    return new Schema(schema);
  }

  /**
   * Tries to log in the user on the configured CubicWeb instance with the given credentials.
   *
   * @param login The user login
   * @param password THe user password
   * @returns A promise returning true if the login was successful, false otherwise
   */
  async login(login: string, password: string) {
    return await this.api.login(login, password);
  }

  /**
   * Get the user associated to this client
   *
   * @returns A promise resolving to the current user
   */
  async currentUser(): Promise<CurrentUser> {
    return await this.api.currentUser();
  }

  /**
   * @param transaction Transaction to execute
   * @returns A promise fulfilled when the transaction is completed
   */
  async executeTransaction(
    transaction: Transaction
  ): Promise<TransactionResult> {
    return await this.transactionBackend.execute(transaction);
  }
}
