import {
  ApiErrorResponse,
  Client,
  RQLParams,
  ResultSet,
} from "@cubicweb/client";
import HttpError from "../HttpError";

/**
 * Converts a CubicWeb error response to a react-admin compatible error object
 *
 * @param error The CubicWeb error response payload
 * @returns
 */
export const apiErrorToReactAdmin = (error: ApiErrorResponse) => {
  return new HttpError(`${error.title}: ${error.message}`, error.status, error);
};

/**
 * Executes the given request and returns the result in a promise.
 * In case of api errors, the promise is rejected with a react-admin http error.
 *
 * @param client The client used to make the request
 * @param requestStr The request string
 * @param params Request parameters
 * @returns
 */
export const executeApiRequest = (
  client: Client,
  requestStr: string,
  params?: RQLParams
) => {
  return new Promise(
    (
      resolve: (value: ResultSet) => void,
      reject: (reason: HttpError) => void
    ) => {
      client
        .execute(requestStr, params)
        .then(resolve)
        .catch((e) => {
          reject(apiErrorToReactAdmin(e));
        });
    }
  );
};

/**
 * Makes a request to the API to get the total number of entities of the given resource
 * matching the given filters
 *
 * @param client The client used to make the request
 * @param resource The resource to fetch data for
 * @param filterRestrictions The filter restrictions in use
 * @returns
 */
export const getTotalEntityCount = async (
  client: Client,
  resource: string,
  filterRestrictions: string[]
) => {
  return await client
    .execute(
      `Any Count(X) WHERE ${[`X is ${resource}`, ...filterRestrictions].join(
        ", "
      )}`
    )
    .then((rows) => rows[0][0] as number);
};
